"""

== MULTIRECLASS TOOL ==
Code: Hans Roelofsen
Ideas: Henk Meeuwsen, Bart de Knegt, Hans Roelofsen
Wageningen Environmental Research Institute
18 augustus 2020

Classes and functions in this script are designed to interact with a Combined Raster (combi_raster output
of the ArcGIS Pro Combine Tool https://pro.arcgis.com/en/pro-app/tool-reference/spatial-analyst/combine.htm). The
combined raster has unique integer value assigned to each unique combination of source rasters. Each source raster is
represented by a field in the output raster attribute table (RAT).

There are two components to the interaction with the combined raster
1. UPSTREAM
New values and categories are assigned to specific value-combinations in the source-rasters, based on 'reclass rules'.
This is the Multireclass part and results in 2 new columns to the RAT: newval and newcat

2. DOWNSTREAM
In downstream, newvals are relabeled and/or aggregated to a new set of categories corresponding to the input need of a
certain geo-spatial model. A new column is added to the RAT for each geo-spatial model showing corresponding categories
for each newval.

Upstream reclass rules, as well as downstream reshuffling are user-provided in an Excel-sheet.
Open the combi_raster in ArcGIS pro to see new_cat and geospatial model classifications.
"""

import os
import pandas as pd
import geopandas as gp
import rasterio as rio
import numpy as np
import datetime
import xlrd
import itertools
import mrt_helpers as mrt
# from sample import mrt_helpers as mrt


class Rule:
    """
    A reclass rule to map applicable values from the Combined raster into newval.
    """
    def __init__(self):
        self.name = None
        self.newval = None # Nevalue
        self.queries = []  # list of query componenents
        self.query = None  # fully parsed query to apply to self.combi_raster.rat
        self.category_desc = None
        self.priority = None
        self.apply = False  # Boolean, should rule be applied
        self.join_type = None  # rule is either strict (AND) or loose (OR)
        self.comment = None
        self.vat_match = None
        self.gross_match = 0
        self.net_match = 0
        self.difference = 0


class Model:
    """
    Class holding all information related to a Downstream model. Terminology:
    outval: int, the numeric value in the output raster.
    cat: str, the description of a land use category related to an outval
    newval: int, the *new value* that is assigned to each pixel in the combined raster as a result of
                  applying the reclass rules.
    combival: int, the values of the combined raster, ie the Value column of the combined raster RAT

    """
    def __init__(self, name, v):


        self.name = name  # name of the model, ie column in downstream sheet
        self.outval2cat = {}  # mapping between Output value and Category description
        self.str_outval2cat = {}  # as above with values as strings (for GTiff tags)
        self.cat2outval = {}  # reverse of outval2cat
        self.newval2cat = {}  # mapping between the Newval and the output category
        self.combival2cat = {}  # mapping between the Value of the Combi raster to the Model category
        self.combival2outval = {} # mapping between the Value of the Combi raster to the model output value
        self.verbose = v

    def get_details(self, xls):
        """
        Read XLS sheet to get details on output values and output categories.
        Assumes XLS has sheet called self.name
        :param xls: adress to XLS
        :return: updated self.outval2cat, self.cat2outval, self.str_outval2cat
        """
        model_df = pd.read_excel(xls, sheet_name=self.name)
        self.outval2cat = dict(zip(model_df.value, model_df.description))
        self.str_outval2cat = dict(zip(model_df.value.astype(str), model_df.description))
        self.cat2outval = dict(zip(model_df.description, model_df.value))

        # update for NoData inoutput: value 0 and category 'geen'
        self.cat2outval['geen'] = 0
        self.outval2cat[0] = 'geen'
        self.str_outval2cat['0'] = 'geen'
        if self.verbose:
            print('created model {}'.format(self.name))


class ReClass:
    """
    Class for reclassing a Combined raster into one or more new categorisations based on reclass rules.

    The reclass module parses reclass rules from a source XLS and and stores as Rule() attributes.


    """

    def __init__(self, xls_dir, xls_in, upstream_sheet, combined_raster_dir, combined_raster_in, downstream_sheet,
                 verbose, base_out=None):
        """
        Initialisation of the whole thing
        :param xls_dir: directory of steering XLS
        :param xls_in: name and extension of steering XLS
        :param upstream_sheet: sheet name with upstream information
        :param downstream_sheet: sheet name with downstream informatio (default None)
        :param combined_raster_dir: directory containing combined raster
        :param combined_raster_in: name and extension of combined raster.
        """
        if not os.path.isfile(os.path.join(xls_dir, xls_in)):
            raise Exception('  Input file {0} not found in directory {1}'.format(xls_in, xls_dir))
        self.__xls = os.path.join(xls_dir, xls_in)
        self.__xls_time = os.path.getmtime(self.__xls)
        self.us_sheet = upstream_sheet
        self.ds_sheet = downstream_sheet
        self.models = None
        self.not_classified = False
        self.rules = []
        self.not_classified_lst = []
        self.written_rasters = []
        self.verbose = verbose
        self.base_out = base_out
        self.combi_raster = CombinedRaster(combined_raster_dir, combined_raster_in, verbose=self.verbose)
        self.timestamp_full = datetime.datetime.now().strftime('%d-%b-%Y_%H:%M:%S')
        self.xls_timestamp = datetime.datetime.fromtimestamp(self.__xls_time).strftime('%Y%m%d-%H%M%S')
        self.xls_timestamp_full = datetime.datetime.fromtimestamp(self.__xls_time).strftime('%d-%b-%Y_%H:%M:%S')
        self.project_name = '{0}-{1}'.format(os.path.splitext(xls_in)[0], self.xls_timestamp)

        try:
            self.rules_df = pd.read_excel(self.__xls, sheet_name=self.us_sheet)
        except (ValueError, xlrd.biffh.XLRDError) as e:
            print('incorrect Excel sheet name: {0}'.format(self.us_sheet))
            raise

        # Identify source rasters from upstream sheet (rules_df) and parse to lowercase
        not_source_rasters = ['newval', 'description', 'remark', 'reclass', 'newvalue', 'new_val', 'join_type']
        self.source_rasters = [x for x in list(self.rules_df) if x.lower() not in not_source_rasters]
        col_rename_dict = dict(zip([x for x in self.rules_df.columns if x not in self.source_rasters],
                                   [x.lower() for x in self.rules_df.columns if x not in self.source_rasters]))
        self.rules_df.rename(columns=col_rename_dict, inplace=True)

        if self.ds_sheet:
            try:
                self.downstream_df = pd.read_excel(self.__xls, sheet_name=self.ds_sheet)
            except (ValueError, xlrd.biffh.XLRDError) as e:
                print('incorrect Excel sheet name: {0}'.format(self.ds_sheet))
                raise
            self.models = [x for x in list(self.downstream_df) if x.lower() not in ['description', 'newval']]
            pd.testing.assert_series_equal(left=self.rules_df.newval, right=self.downstream_df.newval), "Difference detected between upstream and downstream sheets"

            if self.verbose:
                print('Models are: ')
                print('\n'.join(self.models))

        if self.models:
            # Create model objects and attach as attributes
            fails = []
            for model_name in self.models:
                print(model_name)
                try:
                    model_obj = Model(name=model_name, v=self.verbose)
                    model_obj.get_details(xls=self.__xls)
                    setattr(self, model_name, model_obj)
                except (ValueError, xlrd.biffh.XLRDError):
                    print('  no mapping found in Excel sheet for model {}'.format(model_name))
                    fails.append(model_name)
            if fails:
                self.models = [model for model in self.models if model not in fails]

        # Check if col names and values in rules dataframe correspond to provided combined raster
        if not set(self.source_rasters).issubset(set(self.combi_raster.source_rasters)):
            print("Some reclass columns do not match Combined Raster: ")
            print("Columns in the reclass sheet are:")
            print('  \n'.join(self.source_rasters))
            print("\nColumns in the combi raster sheet are:")
            print('  \n'.join(self.combi_raster.source_rasters))

        # Checks for each source raster
        for source_rast in self.source_rasters:
            # the provided (ie asked for) values in a source raster are:
            val_lists = [x for x in self.rules_df.loc[:, source_rast].apply(mrt.parse_rule, negate=False) if x]
            provided_vals = set([item for sublist in val_lists for item in sublist])
            # the valid values are:
            valid_vals = getattr(getattr(self.combi_raster, source_rast), 'vals')

            # Are values provided in the queries actually present in the source raster?
            assert provided_vals.issubset(valid_vals), \
                'Invalid values found for column {0}: {1}'.format(source_rast, provided_vals.difference(valid_vals))

            # Are all values in the source raster actually employed in a query?
            if not valid_vals.issubset(provided_vals):
                self.not_classified = True
                valid_not_used = pd.Series(data=list(valid_vals.difference(provided_vals)), name=source_rast)
                self.not_classified_lst.append(valid_not_used)

        if self.not_classified and self.verbose:
            print('\nWarning, not all valid values for are employed in a Reclass Rule')
            for s in self.not_classified_lst:
                print("{}".format(s.name))
                print('\n'.join(['  {}'.format(str(x)) for x in s]))

    def parse_rules(self):
        """
        Convert the raw information from the rules dataframe to actual queries.
        Note that rule information is stored in both self.rule_df as well as Rule objects stored as attributes and
        in self.rules=[]. It is probably possible to simplify this.
        :return: update self.rule_df and self.rules=[Rule0, Rule1 etc]
        """

        # Create dummy rule (Rule999) to catch all values in the combined raster RAT.
        # For dict merge: https://stackoverflow.com/questions/38987/
        # how-do-i-merge-two-dictionaries-in-a-single-expression-in-python-taking-union-o
        data_p1 = dict.fromkeys(self.source_rasters, 'n')
        data_p2 = {'newval': 0, 'reclass': True, 'description': 'No other rule applies', 'join_type': 'and'}
        dummy_rule = pd.Series(data={**data_p1, **data_p2}, name=999)
        self.rules_df = self.rules_df.append(dummy_rule)

        # Dict to map user provided information to AND or OR join type for rule
        join_types = {'and': '&', 'or': '|', 'en': '&', 'of': '|', 'And': '&', 'Or': '|', 'AND': '&', 'OR': '|',
                      'En': '&', 'Of': '|', 'EN': '&', 'OF': '|'}

        for row in self.rules_df.iterrows():
            series = row[1]
            rule = Rule()
            rule.newval = series.loc['newval']
            rule.category_desc = series.loc['description']
            rule.priority = row[0]
            rule.apply = True if series.loc['reclass'] == 1 else False
            rule.comment = series.loc['remark']  # idem
            rule.name = 'rule{:03}'.format(row[0])
            try:
                rule.join_type = join_types[series.loc['join_type']]
            except KeyError:
                print('Invalid join type indicator "{0}" for rule {1}'.format(series.loc['join_type'], row[0]))
                raise

            # Build query based on the dataframe contents for each source raster
            for reclass_rast in self.source_rasters:
                reclass_rast_query = mrt.parse_rule(series.loc[reclass_rast], getattr(self.combi_raster, reclass_rast))
                if reclass_rast_query:
                    rule.queries.append('{0}.isin({1})'.format(reclass_rast, reclass_rast_query))
            if rule.queries:
                # Build query using the rule join type as connector between the individual queries.
                rule.query = ' {} '.format(rule.join_type).join(rule.queries)
            else:
                rule.apply = False  # never apply a rule that does not have a valid query

            # Append rule.name to list of rule names
            self.rules.append(rule.name)

            # Append as attribute known as rule.name
            # possible use: newvals = [getattr(x, 'newval') for x in self.rules if getattr(x, 'apply')]
            setattr(self, rule.name, rule)

        if len(self.rules) > 254:
            raise Exception('Max 254 reclass rules currently permitted. Sorry...')

        # Set a query for Rule999 that catches all rows of the RAT.
        # https://stackoverflow.com/questions/46822423/
        # pandas-dataframe-query-expression-that-returns-all-rows-by-default
        setattr(getattr(self, "rule999"), 'query', 'tuple()')
        setattr(getattr(self, "rule999"), 'apply', True)

        # Determine which rows in the combi_raster RAT match the query of each rule and calculate coresponding hectare
        for rule_name in [x for x in self.rules if getattr(getattr(self, x), 'query')]:
            vat_match = self.combi_raster.rat.query(getattr(getattr(self, rule_name), 'query')).index
            gross_match = np.multiply(self.combi_raster.rat.loc[vat_match].Count.sum(), self.combi_raster.pixel_area_ha)
            setattr(getattr(self, rule_name), 'vat_match', vat_match)
            setattr(getattr(self, rule_name), 'gross_match', gross_match)

    def look_upstream(self):
        """
        Apply reclass rules to combi raster VAT.
        Warning: currently assumed each rule points to a unique newval.
        If this is not True, note that the upstream df would have > rows than downstream df!
        :return: updated self.combi_raster.rat with 'first_rule (applicable rule), newval and newcat as new columns
        """

        # Read and understand the Reclass rules.
        self.parse_rules()

        # For each value in the combi_raster RAT, determine if each rule is applicable (set 1) or not (set 0)
        for rule_name in self.rules:
            rule = getattr(self, rule_name)
            # only apply rule when indicated as such.
            if rule.apply:  # do only when rule must be applied
                # tag rows complying to rules with value 1
                self.combi_raster.rat.loc[rule.vat_match, rule.name] = 1
                # tag non-complying rules with value 0
                self.combi_raster.rat.loc[self.combi_raster.rat.index.difference(rule.vat_match), rule.name] = 0

        # Determine which Rule is first applicable to each value in the RAT. Then get corresponding newval_val and
        # newval_cat
        self.combi_raster.rule_cols = [col for col in list(self.combi_raster.rat) if col.startswith('rule')]
        self.combi_raster.rat['first_rule'] = self.combi_raster.rat.loc[:, self.combi_raster.rule_cols].idxmax(axis=1)
        self.combi_raster.rat['newval_val'] = \
            self.combi_raster.rat.apply(lambda row: getattr(getattr(self, row.first_rule), 'newval'), axis=1)
        self.combi_raster.rat['newval_cat'] = self.combi_raster.rat.apply(lambda row:
                                                                           getattr(getattr(self, row.first_rule),
                                                                                   'category_desc'), axis=1)

        # Update the rules with the NET area complying to the rule, where rule.apply = True
        for rule_name in [x for x in self.rules if getattr(getattr(self, x), 'apply')]:
            net = np.multiply(self.combi_raster.rat.loc[self.combi_raster.rat.first_rule == rule_name].Count.sum(),
                              self.combi_raster.pixel_area_ha)
            setattr(getattr(self, rule_name), 'net_match', net)

        # Update the rules with the difference between GROSS and NET area. Assert gross >= net
        for rule_name in self.rules:
            net = getattr(getattr(self, rule_name), 'net_match')
            diff = np.subtract(getattr(getattr(self, rule_name), 'gross_match'), net)
            assert diff >= 0, 'Nett area ({0}ha) for {1} exceeds gross area ({2}ha) with {3}ha'.format(
                net, rule_name, getattr(getattr(self, rule_name), 'gross_match'), diff)
            setattr(getattr(self, rule_name), 'difference', diff)

        # Warning if Rule999 applies to any other combinations than 0, 0, 0
        rule999_matches = self.combi_raster.rat.loc[self.combi_raster.rat.first_rule == 'rule999'].index
        if len(rule999_matches) >= 1 and self.verbose:
            print('Warning, some categories have not been reclassed into a new category. '
                  'Paste to Excel for details')
            self.combi_raster.rat.loc[rule999_matches].drop(labels=self.combi_raster.rule_cols, axis=1) \
                .to_clipboard(sep=';', index=False)

        # Add mapping between newval value --> category and category --> value.
        # Note how 'newval' is considered one of several downstream models to ensure compatability with write_raster
        # Also note how Rule999 is integrated here into the newvals, even thought it is not presnet in self.us_sheet
        newval2newcat = dict(zip([getattr(getattr(self, x), 'newval') for x in self.rules if getattr(getattr(self, x), 'apply')],
                                 [getattr(getattr(self, x), 'category_desc') for x in self.rules if getattr(getattr(self, x), 'apply')]))
        newcat2newval = dict(zip([getattr(getattr(self, x), 'category_desc') for x in self.rules if getattr(getattr(self, x), 'apply')],
                                 [getattr(getattr(self, x), 'newval') for x in self.rules if getattr(getattr(self, x), 'apply')]))
        setattr(self, 'newval_val2cat', newval2newcat)
        setattr(self, 'newval_cat2val', newcat2newval)

    def look_downstream(self):
        """
        Add model categorisations and model values to the combined raster RAT
        :return: update self.combi_raster.vat with {model}_cat and {model}_val for model in self.models
        """

        assert hasattr(self.combi_raster.rat, 'newval_val'), 'Look upstream before looking downstream'

        if self.models:
            for model_name in self.models:

                # Mapping between newval and the model output category. Note NAs in model output categorie column
                # because not all newvals are noodzakelijk translated into a output category. Fill NAs with 'geen'
                newval2cat= dict(zip(self.downstream_df.newval, self.downstream_df.loc[:, model_name].fillna('geen')))
                newval2cat[0] = 'geen'  # to catch Rule999
                setattr(getattr(self, model_name), 'newval2cat', newval2cat)  # update newval2cat attr for all models

                # Update combi_raster.rat with model Category and Output value for value in the RAT
                outcat_colname = '{}_cat'.format(model_name)
                outval_colname = '{}_val'.format(model_name)
                self.combi_raster.rat.loc[:, outcat_colname] = \
                    self.combi_raster.rat.newval_val.map(getattr(getattr(self, model_name), 'newval2cat'))
                self.combi_raster.rat.loc[:, outval_colname] = \
                    self.combi_raster.rat.loc[:, outcat_colname].map(getattr(getattr(self, model_name), 'cat2outval'))

                # Mapping between Combiraster RAT value and model output category
                combival2outcat = dict(zip(self.combi_raster.rat.Value, self.combi_raster.rat.loc[:, outcat_colname]))
                setattr(getattr(self, model_name), 'combival2cat', combival2outcat)

                # Mapping between Comiraster RAT Value and model output value
                combival2outval = dict(zip(self.combi_raster.rat.Value, self.combi_raster.rat.loc[:, outval_colname].fillna(0)))
                setattr(getattr(self, model_name), 'combival2outval', combival2outval)


        else:
            print('  no downstream information provided.')

    def cross_tab(self):
        """"
        Cross-tabulation of Rules and matching gross areas
        """
        rule2rule = itertools.product(self.rules, self.rules)
        cross_tab = pd.DataFrame(data=np.zeros(shape=(len(self.rules), len(self.rules))),
                                 columns=self.rules, index=self.rules)
        for a, b in rule2rule:
            sel = self.combi_raster.rat.query('({0} == 1) & ({1} == 1)'.format(a, b))
            if sel.empty:
                if self.verbose:
                    print('{0} & {1}: empty'.format(a, b))
                continue
            else:
                if self.verbose:
                    print('{0} & {1}: {2}'.format(a, b, sel.Count.sum()))
                cross_tab.loc[a, b] = sel.Count.sum()
        setattr(self, 'rules_cross_tab', cross_tab)

    def write_report(self):
        """
        Write summary area statistics for both source rasters, new_cats, downstream models and rules
        :return: Excel file to file
        """

        assert os.path.isdir(self.base_out), 'Invalid output directory'
        excel_out = '{0}_report.xlsx'.format(self.project_name)
        with pd.ExcelWriter(os.path.join(self.base_out, excel_out), mode='w') as writer:

            # Report on input
            metadata = {'input_XLS': self.__xls,
                        'input_XLS_last_modified': self.xls_timestamp_full,
                        'input_downstream_sheet': self.ds_sheet,
                        'input_upstream_sheet': self.us_sheet,
                        'input_combined_raster': self.combi_raster.raster.name,
                        'report creation_date': self.timestamp_full,
                        'created_by': os.getenv('USERNAME')}
            if len(self.written_rasters) > 0:
                for i, j in enumerate(self.written_rasters, start=1):
                    metadata.update({'output_raster_{}'.format(i): j})
            input_df = pd.DataFrame(data=metadata, index=[0])
            input_df.T.to_excel(writer, sheet_name='Metadata', index=True)

            # Report on Rules. Warning: assumes self.rules is ordered in same order as self.rules_df.
            self.rules_df['rule_name'] = [getattr(getattr(self, rule), 'name') for rule in self.rules]
            self.rules_df['query'] = [getattr(getattr(self, rule), 'query') for rule in self.rules]
            self.rules_df['gross_match_ha'] = [getattr(getattr(self, rule), 'gross_match') for rule in self.rules]
            self.rules_df['net_match_ha'] = [getattr(getattr(self, rule), 'net_match') for rule in self.rules]
            self.rules_df['diff_ha'] = [getattr(getattr(self, rule), 'difference') for rule in self.rules]
            self.rules_df.to_excel(writer, sheet_name='ReclassRules', index=False, freeze_panes=(1, 2))

            # Write downstream table for applied rules only
            if self.models:
                applied_newvals = [getattr(getattr(self, rule), 'newval') for rule in self.rules if
                                   getattr(getattr(self, rule), 'apply')]
                self.downstream_df.loc[self.downstream_df.newval.isin(applied_newvals), :] \
                    .to_excel(writer, sheet_name='downstream', index=False, freeze_panes=(1,2))

            # Report on combi raster RAT
            self.combi_raster.rat.drop(labels=self.combi_raster.rule_cols + ['geometry'], axis=1) \
                .sort_values(by='first_rule') \
                .to_excel(writer, sheet_name='Reclassed', index=False, freeze_panes=(2, 1))

            # Report on values in source raster that are not classified
            if self.not_classified:
                pd.concat(self.not_classified_lst, axis=1).to_excel(writer, sheet_name='NotClassified', index=False,
                                                                    freeze_panes=(2, 1))

            # Report on cross-tabulation
            if hasattr(self, 'rules_cross_tab'):
                self.rules_cross_tab.to_excel(writer, sheet_name='Kruistabel', index=True, freeze_panes=(1, 1))

            # Report on source rasters
            for src_raster in self.source_rasters:
                stats = pd.pivot_table(data=self.combi_raster.rat, index=src_raster, values='Count', aggfunc='sum')
                stats.rename(columns={'Count': 'pixel_count'}, inplace=True)
                stats['hectare'] = stats.pixel_count.multiply(self.combi_raster.pixel_area_ha)
                stats.index.name = 'Pixel Value {0}'.format(src_raster)
                stats.to_excel(writer, sheet_name='source_raster_{0}'.format(src_raster))

            # Report on downstream models
            if self.models:
                for model in self.models:
                    stats = pd.pivot_table(data=self.combi_raster.rat, index='{}_cat'.format(model), values='Count',
                                           aggfunc='sum')
                    stats.rename(columns={'Count': 'pixel_count'}, inplace=True)
                    stats['hectare'] = stats.pixel_count.multiply(self.combi_raster.pixel_area_ha)
                    # stats['value'] = stats.index.map()
                    stats.index.name = 'Category {0}'.format(model)
                    stats.to_excel(writer, sheet_name='ESDModel_{0}'.format(model))


    def write_raster(self, request_raster, extent='BNL', out_format='geotiff'):
        """
        Write a new raster with pixel values reclass to one of the
        columns of self.combi_raster.rat and optionally cropped to either the LCEU or BNL format
        :param request_raster: name of the RAT attribute table column to which to map pixel values
        :param extent: profile for extent of output raster {BNL, LCEU}. Default: BNL
        :out_format: geospatial raster file format {geotiff, flt, map}. Default: geotiff
        :return: TIF file in out_dir
        """

        assert os.path.isdir(self.base_out), 'Invalid output directory'
        assert request_raster in self.source_rasters + ['newval', 'Value'] + (self.models if self.models else None), 'Invalid request'

        # Dictionary to map between combiraster values and requested values
        value2val = getattr(getattr(self, request_raster), 'combival2outval')
        out_dtype = np.uint8 if max([v for _, v in value2val.items()]) <= 255 else np.uint16
        out_nodata = 255 if out_dtype == np.uint8 else 65535

        if self.combi_raster.raster.nodata:
            value2val[self.combi_raster.raster.nodata] = out_nodata

        # Fetch raster profile with extent and pixel size
        prof = mrt.raster_profile(extent=extent, pxl_size=self.combi_raster.raster.res[0])
        prof.update(datatype=out_dtype, nodata=out_nodata)
        prof.update(driver=mrt.format2driver(out_format))
        out_name = '{0}_{1}'.format(self.project_name, request_raster)

        # Write to new file, one block at the time. Optionally reduce extent from BNL extent to LCEU extent
        #  BNL  (     0, 300.000, 280.000, 625.000)
        #  LCEU (10.000, 300.000, 280.000, 620.000)
        # Crop top rows: (625.000-620.000)/2.5 = 2.000
        # Crop left cols: (10.000 - 0)/2.5 = 4.000

        destination = os.path.join(self.base_out, '{0}.{1}'.format(out_name, mrt.format2ext(out_format)))
        with rio.open(destination, 'w', **prof) as dest:
            dest.update_tags(creation_date=self.timestamp_full,
                             created_by=os.getenv('USERNAME'),
                             src_file=self.combi_raster.raster.name,
                             content=request_raster,
                             remap_xls=self.project_name,
                             spatial_extent=extent)
            # Hard-code all pixel values and corresponding categories into the raster
            dest.update_tags(**getattr(getattr(self, request_raster), 'str_outval2cat'))

            src_blocks = [x for x in self.combi_raster.raster.block_windows(1)]
            for block in mrt.progress_bar(iterable=src_blocks, prefix='Reclass to {0}'.format(request_raster),
                                          suffix='Complete', length=50):

                # block contents is ((row, col), ((row_start, row_end), (col_start, col_end)))
                # See: https://rasterio.readthedocs.io/en/latest/topics/windowed-rw.html#blocks
                row, col = block[0][0], block[0][1]
                row_start, row_end = block[1][0][0], block[1][0][1]
                col_start, col_end = block[1][1][0], block[1][1][1]

                # check that blocks are equal to 1 row of the combi_raster, which happens to be the case for AGPro
                # Combine tool outputs, but may or may not be so for other rasters. If they aren't, it will be harder
                # to apply extent cropping.
                assert np.subtract(row_end, row_start) == 1, "Block shapes are not to expectations"
                assert np.subtract(col_end, col_start) == self.combi_raster.raster.shape[1]

                if extent == 'LCEU':
                    pxl_top_crop = np.divide(np.subtract(625000, 620000), self.combi_raster.raster.res[0])
                    pxl_left_crop = np.divide(np.subtract(10000, 0), self.combi_raster.raster.res[0])

                if extent == 'LCEU' and row_start < pxl_top_crop:
                    continue  # skipping top rows
                elif extent == 'LCEU' and row_start >= pxl_top_crop:
                    read_window = ((row_start, row_end), (pxl_left_crop, col_end))
                    write_window = ((row_start - pxl_top_crop, row_end - pxl_top_crop), (0, col_end - pxl_left_crop))
                else:
                    read_window = block[1]
                    write_window = block[1]

                # Read portion of combi raster
                arr = self.combi_raster.raster.read(1, window=read_window)
                
                # Reclass according to dictionary 
                arr_reclass = mrt.vec_translate(arr, value2val).astype(out_dtype)

                # Write portion to file
                dest.write(arr_reclass, window=write_window, indexes=1)

        if self.models:
            # Write DBF to file with Value=Unique values of the request raster and Description=Category Description
            out_vat = gp.GeoDataFrame.from_dict(getattr(getattr(self, request_raster), 'outval2cat'),
                                                orient='index', columns=['Description'])
            out_vat.loc[:, 'Value'] = out_vat.index
            out_vat.loc[:, 'geometry'] = None
            mrt.write_vat(gdf=out_vat, out_dir=self.base_out, out_name=out_name, v=self.verbose)
    
            # Write QGIS Layer Style File (QML) to file
            mrt.write_qml(tab=out_vat, out_dir=self.base_out, out_name=out_name)

        # assert everything has worked and add to list of written rasters
        if os.path.isfile(destination):
            self.written_rasters.append(destination)

    def test(self):
        """"Test integrity of all inputs and report possibilities."""
        print('\n\nReclass objects based on XLS {0} and Combi Raster {1} with {2} rules for the following source rasters:'
              .format(self.__xls, self.combi_raster.raster.name, len(self.rules)))
        for i, rast in enumerate(self.combi_raster.source_rasters, start=1):
            print('  {0}: {1}'.format(i, rast))
        print('Netto area in ha for each rule:')
        for rule_name in self.rules:
            print('{}: {:10.1f} ha'.format(getattr(getattr(self, rule_name), 'name'),
                                           getattr(getattr(self, rule_name), 'net_match')))

        if self.models:
            print('\nCan create input for following downstream categories:')
            for i, model in enumerate(self.models + ['newval'], start=1):
                print('  {0}: {1}'.format(i, model))
        if self.base_out:
            print('\n{0} {1} a valid output directory'.format(self.base_out,
                                                              'is' if os.path.isdir(self.base_out) else 'is not'))


class CombinedRaster:
    """
    Class holding all information related to a integer raster + RAT resulting from arcpy.combine
    See: https://desktop.arcgis.com/en/arcmap/10.3/tools/spatial-analyst-toolbox/combine.htm
    """

    def __init__(self, rast_dir, rast_in, verbose):
        """
        Store info related to the raster and read as rasterio object.
        :param rast_dir: directory containing combined raster
        :param rast_in: name and extension of combined raster, should be GDAL format
        """
        assert os.path.isdir(rast_dir), 'invalid directory for combined raster'
        assert os.path.isfile(os.path.join(rast_dir, rast_in)), 'combined raster does not exist'

        self.rast_dir = rast_dir
        self.basename, self.ext = os.path.splitext(rast_in)
        self.raster = rio.open(os.path.join(rast_dir, rast_in))
        try:
            self.rat = gp.read_file(os.path.join(rast_dir, '{}.tif.vat.dbf'.format(self.basename)))
            # AGPro delivers columnnames inconsistently: sometimes Count, sometimes COUNT.
            self.rat.columns = [x.capitalize() if x.lower() in ['count', 'value'] else x for x in list(self.rat)]
        except ValueError:
            print('Raster Attribute Table not found')
            raise

        self.source_rasters = [x for x in list(self.rat) if x not in ['Value', 'Count', 'geometry']]
        self.pxl_area_m2 = np.square(self.raster.res[0])
        self.pixel_area_ha = np.divide(self.pxl_area_m2, 10000)
        self.rule_cols = []
        self.negative_vals_found = False

        for source_raster in self.source_rasters:
            setattr(self, source_raster, SourceRaster(name=source_raster, vals=set(self.rat.loc[:, source_raster])))

        # Check for < 0 values in the Count field, which should not be possible, but appears to happen with AGPro,
        # esp for 2.5 m rasters. Probably due to integer overflow of the Long datatype in the VAT Count field.
        # This assumes that each pixel in the raster is represented in the raster RAT, which MAY NOT BE THE CASE
        # when AGPro NoData masks are present.

        if any(self.rat.Count < 0):
            neg_index = self.rat.loc[self.rat.Count < 0].index
            if len(neg_index) == 1:

                self.negative_vals_found = True
                total_count = self.raster.shape[0] * self.raster.shape[1]  # total nr of pixels in raster
                accounted_for = self.rat.loc[self.rat.index.difference(neg_index)].Count.sum()  # Pixel count of RAT vals w Count > 0
                negative_count = self.rat.at[neg_index.tolist()[0], 'Count']  # Given negative value, incorrect
                should_be = np.subtract(total_count, accounted_for)  # substitue value = total_count - accounted_for

                if verbose:
                    print('  !Warning, negative value found in raster VAT. Substituting {0} for {1}'.
                          format(negative_count, should_be))

                # update the RAT and also record change
                self.rat.loc[neg_index, 'Count'] = should_be
                self.rat.loc[:, 'Count_orig'] = self.rat.Count
                self.rat.loc[neg_index, 'Count_orig'] = negative_count
                assert self.rat.Count.sum() == total_count
            else:
                raise Exception('Multiple negative values detected in raster VAT.')


class SourceRaster:
    """
    Class for holding all information of a source raster incorporated into a CombinedRaster, as derived from the
    CombinedRaster RAT
    """
    def __init__(self, name, vals):
        self.name = name
        self.vals = vals


if __name__ == '__main__':
    bdb = ReClass(xls_dir=r'c:\Users\roelo008\Wageningen University & Research\Natuurverkenning breder doelbereik - General\LUKaart\huidig',
                  xls_in='nvk_huidig.xlsx',
                  upstream_sheet='upstream_250cm',
                  downstream_sheet='downstream',
                  combined_raster_dir=r'w:\PROJECTS\Nvk_bdb\a_geodata\a_huidig\b_combined',
                  combined_raster_in='nvk_combi_20201104.tif',
                  base_out=r'c:\apps\temp_geodata\bdb\z_out\20200929',
                  verbose=False)
    bdb.look_upstream()
    bdb.look_downstream()
    bdb.test()
    # bdb.write_raster(request_raster='newval', out_dir=r'c:\apps\temp_geodata\bdb\z_out\rast')
    # bdb.rewrite_vat(out_dir=r'c:\apps\temp_geodata\bdb\z_out\250cm\v2')
    # bdb.write_report()
    # bdb.write_raster(request_raster='LCEU', extent='LCEU', vat=True)

'''
python MRT.py W:\PROJECTS\EcoSysDiensten_WEnR\b_data\c_ScenarioSpecifiek\BDB_Basiskaart\testing03\BNL_rules_reclass_BDB_HDR.xlsx c:\apps\temp_geodata\bdb\a_combined\10m\bdb_combined_20200826.tif upstream_10m --test

'''
# from sample.z_utils import mrt_helpers as mrt
'''
combined_raster_dir = r'c:\apps\temp_geodata\bdb\a_combined\10m'
combined_raster_in = 'bdb_combined_20200826.tif'
reclass_dict = dict(zip(bdb.combi_raster.rat.Value, bdb.combi_raster.rat.LCEU_val))
reclass_dict[0] = 0

src = rio.open(os.path.join(combined_raster_dir, combined_raster_in))
prof = src.profile
prof.update(dtype=rio.int32)
dst = rio.open(r'c:/apps/temp_geodata/test/bdb_reclass.tif', mode='w', **prof)

import time
i = 0
lijstje = [x for x in src.block_windows(1)]
for item in mrt.progressBar(iterable=lijstje, prefix='Progress:', suffix='Complete', length=50):
    window = item[1]
    r = src.read(1, window=window)
        # r_new = mrt.vec_translate(r, reclass_dict)
        # dst.write(r_new, window=window, indexes=1)
src.close()
dst.close()
'''



