"""
Helper functions for the Multireclass functionality
Hans Roelofsen, 24 augustus 2020

"""

import os
import random
import argparse
import rasterio as rio
import numpy as np
import affine
import xml.etree.ElementTree as ET

def vec_translate(a, my_dict):
    """
    Function to apply my_dict to each element in in array a. See:
    https://stackoverflow.com/questions/16992713/translate-every-element-in-numpy-array-according-to-key

    Note that key errors are not caught!
    """
    return np.vectorize(my_dict.__getitem__)(a)


def parse_rule(rule, category=None, negate=True):
    """
    Parse rule information in reclass df to a list of values.

    INPUT    OUTPUT
    1        [1]
    "1"      [1]
    "1,2,3"  [1, 2, 3]
    "-1"     [2, 3]
    "-1,2"   [3]
    "n"      None
    else     Exception

    :param rule: string or int
    :param category: name source raster, used to negate number(s) starting with "-". Default None
    :param negate: return negation of provided numbers if txt.startswith('-'). Default True
    :return: see above.
    """

    if isinstance(rule, int):
        out = [rule]
    elif isinstance(rule, str):
        txt2 = rule.strip()
        if txt2 == 'n':
            out = None
        elif txt2.startswith('-'):
            neg_vals = [int(i) for i in txt2[1:].split(',')] if ',' in txt2 else [int(txt2[1:])]
            try:
                all_vals = getattr(category, 'vals')
            except AttributeError:
                raise Exception('please provide a valid category')
            out = list(all_vals.difference(set(neg_vals))) if negate else list(neg_vals)
        elif ',' in txt2:
            out = [int(i) for i in txt2.split(',')]
        else:
            try:
                out = [int(txt2)]
            except ValueError as e:
                print('Error parsing reclass rule')
                raise
    else:
        raise Exception('unexpected reclass rule found: {0}'.format(rule))
        # out = False
    return out


def raster_profile(extent, pxl_size):
    """
    Return rasterio raster profiles
    :param extent: extent in RD-New. Either LCEU (10.000, 300.000, 280.000, 620.000)
                                            BNL (0, 300.000, 280.000, 625.000)
    :pxl_size: either 10 or 2.5 m.
    :return:raster profile
    """

    default = rio.default_gtiff_profile.defaults
    default.update(width=27000,
                   nodata=255.0,
                   blockxsize=128,
                   blockysize=128,
                   crs=rio.crs.CRS({'lat_0': 52.15616055555555, 'k': 0.9999079, 'y_0': 463000, 'x_0': 155000,
                                    'ellps': 'bessel', 'units': 'm', 'lon_0': 5.38763888888889, 'no_defs': True,
                                    'proj': 'sterea'}),
                   height=32000,
                   affine=affine.Affine(10.0, 0.0, 0, 0.0, -10.0, 625000.0),
                   bigtiff='NO',
                   count=1)

    if extent == 'LCEU' and pxl_size == 2.5:
        default.update(affine=affine.Affine(2.5, 0, 10000, 0, -2.5, 620000),
                       width=108000, height=128000)
    elif extent == 'LCEU' and pxl_size == 10:
        default.update(affine=affine.Affine(10, 0, 10000, 0, -10, 620000),
                       width=27000, height=32000)
    elif extent == 'BNL' and pxl_size == 2.5:
        default.update(affine=affine.Affine(2.5, 0, 0, 0, -2.5, 625000),
                       width=112000, height=130000)
    elif extent == 'BNL' and pxl_size == 10:
        default.update(affine=affine.Affine(10, 0, 0, 0, -10, 625000),
                       width=28000, height=32500)
    else:
        raise Exception('Not a valid raster profile request')
    return default


def format2ext(raster_format):
    """
    Return file extension corresponding to requested geospatial raster file format
    :param raster_format: requested format {geotiff, flt, map}
    :return: {tiff, flt, map}
    """

    driver2ext = {
        'geotiff': 'tiff',
        'flt': 'flt',
        'map': 'map'}

    try:
        return driver2ext[raster_format]
    except KeyError:
        raise Exception('Invalid format: {}'.format(raster_format))


def format2driver(raster_format):
    """
    Return gdal file driver corresponding to requested geospatial raster file format
    :param raster_format: requested format {geotiff, flt, map}
    :return: {GTiff, EHdr, PCRaster}
    """

    format2driver = {
        'geotiff': 'GTiff',
        'flt': 'EHdr',
        'map': 'PCRaster'}

    try:
        return format2driver[raster_format]
    except KeyError:
        raise Exception('Invalid format: {}'.format(raster_format))


def write_vat(gdf, out_dir, out_name, v):
    """
    Write a geodataframe to file as shapefile, then remove all but the DBF file so that the DBF serves as raster VAT
    :param gdf: geopandas geodataframe
    :param out_dir: output dir
    :param out_name: outputname without extension
    :param v: verbose, booelean
    :return: <output_dir>/<output_name>.tif.vat.dbf
    """

    gdf.to_file(os.path.join(out_dir, '{0}.shp'.format(out_name)))

    # delete all shapefile components except dbf
    redundant_shp_files = [f for f in os.listdir(out_dir) if os.path.splitext(f)[1] in ['.shp', '.shx', '.prj',
                                                                                        '.sbn', '.sbx', '.cpg',
                                                                                        'shp.xml']]
    for f in redundant_shp_files:
        os.remove(os.path.join(out_dir, f))

    # rename shapefile DBF for ArcGIS Pro to recognize it as part of the TIFF
    old_name = '{0}.dbf'.format(out_name)
    new_name = '{0}.tif.vat.dbf'.format(out_name)
    os.rename(os.path.join(out_dir, old_name), os.path.join(out_dir, new_name))
    if os.path.isfile(os.path.join(out_dir, new_name)) and v:
        print('Created raster VAT {}'.format(new_name))


def progress_bar(iterable, prefix='', suffix='', decimals=1, length=100, fill='█', printEnd="\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    credits: https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
    """
    total = len(iterable)
    # Progress Bar Printing Function

    def printProgressBar(iteration):
        percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + '-' * (length - filledLength)
        print('\r{0} |{1}| {2}% {3}'.format(prefix, bar, percent, suffix), end=printEnd)

    # Initial Call
    printProgressBar(0)

    # Update Progress Bar
    for i, item in enumerate(iterable):
        yield item
        printProgressBar(i + 1)
    # Print New Line on Complete
    print()


def geospatial_raster(x):
    """
    Is x a geospatial raster?
    :param x: file path
    :return: x if True else argparse.ArgumentTypeError
    """

    if os.path.isfile(x):
        try:
            rast = rio.open(x)
            return x
        except(rio.errors.RasterioIOError, rio.errors.FileError, rio._err.CPLE_OpenFailed) as e:
            raise argparse.ArgumentTypeError(e)
    else:
        raise argparse.ArgumentTypeError('Not a file: {0}'.format(x))


def rand_web_color_hex():
    """"
    Generate random color.
    https://blog.stigok.com/2019/06/22/generate-random-color-codes-python.html
    """

    rgb = ""
    for _ in "RGB":
        i = random.randrange(0, 2**8)
        rgb += i.to_bytes(1, "big").hex()
    return '#{}'.format(rgb)


def write_qml(tab, out_dir, out_name):
    """
    Write QGIS QML file
    :param tab:
    :param out_dir:
    :param out_name:
    :return:
    """

    tree = ET.parse(r'c:\Users\roelo008\OneDrive - WageningenUR\a_projects\NVK_BDB\bdb_huidig_qgis_style_file.qml')
    root = tree.getroot()
    target = root[2][0][2]
    assert target.tag == 'colorPalette'
    for x in target.findall('paletteEntry'):
        target.remove(x)

    vals = tab.Value
    desc = tab.Description
    for v, d in zip(vals, desc):
        ET.SubElement(target, 'paletteEntry', attrib={'color': rand_web_color_hex(), 'value': str(v),
                                                      'label': '{0}: {1}'.format(v, d), 'alpha': "255"})
    tree.write(os.path.join(out_dir, '{}.qml'.format(out_name)))


def array2val(arr, method, dt, prio=None, th=None, ignore=None):
    """
    Parse a Numpy array to a single output value based on method. Usefull for aggregation
    :param arr: 2D numpy array
    :param method: [sum, min, max, median, henk]. Henk means to give priority to a value if it occurs >= a treshold
    :param dt: datatype of output array
    :param prio: the value to be prioritized if method == henk
    :param th: occurence treshold for the priority value. Return median if treshold is not exceeded
    :param ignore: value to ignore in all methods
    :return: 2D numpy array with shape (1,1) with value following above, unless the input array has all same values,
             then return that value. This trumps ignore values
    """

    # All values are the same, return this value
    if arr.std() == 0:
        return np.array([[arr[0, 0]]]).astype(dt)

    # Mask away ignored values if requested
    if ignore is not None:
        arr = np.ma.array(arr, mask=np.where(arr == ignore, True, False))
        v, c = np.unique(arr, return_counts=True)
        vals = v.data[~v.mask]  # Values with ignore value removed
        counts = c[~v.mask]     # Counts with ignore value removed
    else:
        vals, counts = np.unique(arr, return_counts=True)

    if method == 'median':
        out = vals[counts.argmax()]
        return np.array([[out]]).astype(dt)

    elif method == 'henk':
        if counts[np.where(vals == prio)] >= th:  # priority value is in the array and exceeds treshold
            return np.array([[prio]]).astype(dt)
        else:  # priority value does not exceed treshold or is not in the array at all.
            out = vals[counts.argmax()]  # default to most occuring value
            return np.array([[out]]).astype(dt)

    elif method == 'sum':
        return np.array([[arr.sum()]]).astype(dt)

    elif method == 'min':
        return np.array([[arr.min()]]).astype(dt)

    elif method == 'max':
        return np.array([[arr.max()]]).astype(dt)

    else:
        raise Exception('Invalid method for aggregation')


